/**********************************************************************
  ߃fR[h  bit0-5=0b111011  bit27-30
  $Id: op_3b_subfunc.c,v 1.2 2002/08/28 17:08:52 tane Exp $
**********************************************************************/
#include <math.h>
#include "common.h"
#include "ppcsim.h"
#include "processor.h"
#include "funcs.h"
#include "memory.h"
#include "op_exec.h"

/*fdivsx       1 1 1 0 1 1 > D     < > A     < > B     < 0 0 0 0 0 1 0 0 1 0 R*/
/*fsubsx       1 1 1 0 1 1 > D     < > A     < > B     < 0 0 0 0 0 1 0 1 0 0 R*/
/*faddsx       1 1 1 0 1 1 > D     < > A     < > B     < 0 0 0 0 0 1 0 1 0 1 R*/
/*fresx(*4)    1 1 1 0 1 1 > D     < 0 0 0 0 0 > B     < 0 0 0 0 0 1 1 0 0 0 R*/
/*fmulsx       1 1 1 0 1 1 > D     < > A     < 0 0 0 0 0 > C     < 1 1 0 0 1 R*/
/*fmsubsx      1 1 1 0 1 1 > D     < > A     < > B     < > C     < 1 1 1 0 0 R*/
/*fmaddsx      1 1 1 0 1 1 > D     < > A     < > B     < > C     < 1 1 1 0 1 R*/
/*fnmsubsx     1 1 1 0 1 1 > D     < > A     < > B     < > C     < 1 1 1 1 0 R*/
/*fnmaddsx     1 1 1 0 1 1 > D     < > A     < > B     < > C     < 1 1 1 1 1 R*/

int op_3b_002_func( struct MACHINE_RESOURCE *mres , int mode )  /* fdivsx */
{
  unsigned int	fd,fa,fb ;
  unsigned int	rd ;

  fd=(mres->inst & 0x03e00000)>>21 ;
  fa=(mres->inst & 0x001f0000)>>16 ;
  fb=(mres->inst & 0x0000f800)>>11 ;
  rd=(mres->inst & 0x00000001)     ;

  if( DO_EXEC_DIS(mode) ){
    dis_print_frfrfr(fd,fa,fb) ;
  }
  
  if( DO_EXEC_INST(mode) ){
    if( rd!=0 ) return(NOT_IMPLEMENTED_OP) ;
    mres->FPR[fd].f=(float)mres->FPR[fa].f / (float)mres->FPR[fb].f ;
    mres->pc += 4 ;
  }
  
  return(0) ;
}


int op_3b_004_func( struct MACHINE_RESOURCE *mres , int mode )  /* fsubsx */
{
  unsigned int	fd,fa,fb ;
  unsigned int	rd ;

  fd=(mres->inst & 0x03e00000)>>21 ;
  fa=(mres->inst & 0x001f0000)>>16 ;
  fb=(mres->inst & 0x0000f800)>>11 ;
  rd=(mres->inst & 0x00000001)     ;

  if( DO_EXEC_DIS(mode) ){
    dis_print_frfrfr(fd,fa,fb) ;
  }
  
  if( DO_EXEC_INST(mode) ){
    if( rd!=0 ) return(NOT_IMPLEMENTED_OP) ;
    mres->FPR[fd].f=(float)mres->FPR[fa].f - (float)mres->FPR[fb].f ;
    mres->pc += 4 ;
  }
  
  return(0) ;
}


int op_3b_005_func( struct MACHINE_RESOURCE *mres , int mode )  /* faddsx */
{
  unsigned int	fd,fa,fb ;
  unsigned int	rd ;

  fd=(mres->inst & 0x03e00000)>>21 ;
  fa=(mres->inst & 0x001f0000)>>16 ;
  fb=(mres->inst & 0x0000f800)>>11 ;
  rd=(mres->inst & 0x00000001)     ;

  if( DO_EXEC_DIS(mode) ){
    dis_print_frfrfr(fd,fa,fb) ;
  }
  
  if( DO_EXEC_INST(mode) ){
    if( rd!=0 ) return(NOT_IMPLEMENTED_OP) ;
    mres->FPR[fd].f=(float)mres->FPR[fa].f + (float)mres->FPR[fb].f ;
    mres->pc += 4 ;
  }
  
  return(0) ;
}


int op_3b_008_func( struct MACHINE_RESOURCE *mres , int mode )  /* fresx */
{
  unsigned int	fd,fb ;
  unsigned int	rd ;

  fd=(mres->inst & 0x03e00000)>>21 ;
  fb=(mres->inst & 0x0000f800)>>11 ;
  rd=(mres->inst & 0x00000001)     ;
  
  if( DO_EXEC_DIS(mode) ){
    dis_print_frfr(fd,fb) ;
  }
  
  if( DO_EXEC_INST(mode) ){
    if( rd!=0 ) return(NOT_IMPLEMENTED_OP) ;
    if( (float)mres->FPR[fb].f==0.0 ){ /**FPSCR[VE],FPSCR[ZE]̏ԂlȂĂ͂ȂȂ**/
      return(DIVIDED_BY_ZERO) ;
    }
    mres->FPR[fd].f=1.0/(float)mres->FPR[fb].f ;
    mres->pc += 4 ;
  }
  
  return(0) ;
}


int op_3b_009_func( struct MACHINE_RESOURCE *mres , int mode )  /* fmulsx */
{
  unsigned int	fd,fa,fc ;
  unsigned int	rd ;

  fd=(mres->inst & 0x03e00000)>>21 ;
  fa=(mres->inst & 0x001f0000)>>16 ;
  fc=(mres->inst & 0x000007c0)>>6  ;
  rd=(mres->inst & 0x00000001)     ;

  if( DO_EXEC_DIS(mode) ){
    dis_print_frfrfr(fd,fa,fc) ;
  }
  
  if( DO_EXEC_INST(mode) ){
    if( rd!=0 ) return(NOT_IMPLEMENTED_OP) ;
    mres->FPR[fd].f=(float)mres->FPR[fa].f * (float)mres->FPR[fc].f ;
    mres->pc += 4 ;
  }
  
  return(0) ;
}
  

int op_3b_00c_func( struct MACHINE_RESOURCE *mres , int mode )  /* fmsubsx */
{
  unsigned int	fd,fa,fb,fc ;
  unsigned int	rd ;

  fd=(mres->inst & 0x03e00000)>>21 ;
  fa=(mres->inst & 0x001f0000)>>16 ;
  fb=(mres->inst & 0x0000f800)>>11 ;
  fc=(mres->inst & 0x000007c0)>>6  ;
  rd=(mres->inst & 0x00000001)     ;

  if( DO_EXEC_DIS(mode) ){
    dis_print_frfrfrfr(fd,fa,fb,fc) ;
  }
  
  if( DO_EXEC_INST(mode) ){
    if( rd!=0 ) return(NOT_IMPLEMENTED_OP) ;
    mres->FPR[fd].f=((float)mres->FPR[fa].f * (float)mres->FPR[fc].f) - (float)mres->FPR[fb].f ;
    mres->pc += 4 ;
  }
  
  return(0) ;
}
  

int op_3b_00d_func( struct MACHINE_RESOURCE *mres , int mode )  /* fmaddsx */
{
  unsigned int	fd,fa,fb,fc ;
  unsigned int	rd ;

  fd=(mres->inst & 0x03e00000)>>21 ;
  fa=(mres->inst & 0x001f0000)>>16 ;
  fb=(mres->inst & 0x0000f800)>>11 ;
  fc=(mres->inst & 0x000007c0)>>6  ;
  rd=(mres->inst & 0x00000001)     ;

  if( DO_EXEC_DIS(mode) ){
    dis_print_frfrfrfr(fd,fa,fb,fc) ;
  }
  
  if( DO_EXEC_INST(mode) ){
    if( rd!=0 ) return(NOT_IMPLEMENTED_OP) ;
    mres->FPR[fd].f=((float)mres->FPR[fa].f * (float)mres->FPR[fc].f) + (float)mres->FPR[fb].f ;
    mres->pc += 4 ;
  }
  
  return(0) ;
}
  

int op_3b_00e_func( struct MACHINE_RESOURCE *mres , int mode )  /* fnmsubsx */
{
  unsigned int	fd,fa,fb,fc ;
  unsigned int	rd ;
  unsigned int *fpr_u ;

  fd=(mres->inst & 0x03e00000)>>21 ;
  fa=(mres->inst & 0x001f0000)>>16 ;
  fb=(mres->inst & 0x0000f800)>>11 ;
  fc=(mres->inst & 0x000007c0)>>6  ;
  rd=(mres->inst & 0x00000001)     ;

  if( DO_EXEC_DIS(mode) ){
    dis_print_frfrfrfr(fd,fa,fb,fc) ;
  }
  
  if( DO_EXEC_INST(mode) ){
    if( rd!=0 ) return(NOT_IMPLEMENTED_OP) ;
    
    mres->FPR[fd].f=((float)mres->FPR[fa].f * (float)mres->FPR[fc].f) - (float)mres->FPR[fb].f ;
    fpr_u = (unsigned int *)&mres->FPR[fd].f ;
#ifdef LITTLE_ENDIAN
    fpr_u = fpr_u+1 ;
    *fpr_u= *fpr_u ^ 0x80000000 ;
#else
    *fpr_u= *fpr_u ^ 0x80000000 ;
#endif
    mres->pc += 4 ;
  }
  
  return(0) ;
}
  

int op_3b_00f_func( struct MACHINE_RESOURCE *mres , int mode )  /* fnmaddsx */
{
  unsigned int	fd,fa,fb,fc ;
  unsigned int	rd ;
  unsigned int *fpr_u ;

  fd=(mres->inst & 0x03e00000)>>21 ;
  fa=(mres->inst & 0x001f0000)>>16 ;
  fb=(mres->inst & 0x0000f800)>>11 ;
  fc=(mres->inst & 0x000007c0)>>6  ;
  rd=(mres->inst & 0x00000001)     ;

  if( DO_EXEC_DIS(mode) ){
    dis_print_frfrfrfr(fd,fa,fb,fc) ;
  }
  
  if( DO_EXEC_INST(mode) ){
    if( rd!=0 ) return(NOT_IMPLEMENTED_OP) ;
    
    mres->FPR[fd].f=((float)mres->FPR[fa].f * (float)mres->FPR[fc].f) + (float)mres->FPR[fb].f ;
    fpr_u = (unsigned int *)&mres->FPR[fd].f ;
#ifdef LITTLE_ENDIAN
    fpr_u = fpr_u+1 ;
    *fpr_u= *fpr_u ^ 0x80000000 ;
#else
    *fpr_u= *fpr_u ^ 0x80000000 ;
#endif
    mres->pc += 4 ;
  }

  return(0) ;
}
